/*******************************************************************
  uLan Communication - uL_DRV - multiplatform uLan driver

  ul_cps1.c	- chip driver for USB2uLan device developed
                  by Petr Smolik

  (C) Copyright 1996-2004 by Pavel Pisa - project originator
        http://cmp.felk.cvut.cz/~pisa
  (C) Copyright 1996-2004 PiKRON Ltd.
        http://www.pikron.com
  (C) Copyright 2002-2004 Petr Smolik
  

  The uLan driver project can be used and distributed 
  in compliance with any of next licenses
   - GPL - GNU Public License
     See file COPYING for details.
   - LGPL - Lesser GNU Public License
   - MPL - Mozilla Public License
   - and other licenses added by project originator

  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  WARNING: if you delete all lines, you are not allowed to
  distribute code or sources in any form.
 *******************************************************************/

/*******************************************************************/
/* usb2ulan device requests */

#define U2U_PKT_BUF_SIZE       32     /* driver maximal supported packed size */
#define U2U_MAX_MSG_INPR        4     /* maximal number of messages in progress */

#define VENDOR_START_ULAN       0     /* Inform converter about uLan driver ready */
#define VENDOR_STOP_ULAN        1     /* Inform about uLan driver stopping  */
#define VENDOR_IS_RUNNING_ULAN  2
#define VENDOR_ARC_ON_ULAN      3     /* Set promiscuous monitoring mode */
#define VENDOR_ARC_OFF_ULAN     4     /* Return to std. adress recognition mode */

#define U2UBFL_NORE     0x040         /* Do not try to repeat if error occurs */ 
#define U2UBFL_TAIL     0x020         /* Message has tail frame */
#define U2UBFL_REC      0x010         /* Request receiption of block */
#define U2UBFL_FAIL     0x008         /* Message cannot be send - error */
#define U2UBFL_PROC	0x004         /* Message succesfull send */
#define U2UBFL_AAP	0x003         /* Request imediate proccessing of frame by receiver station with acknowledge */
#define U2UBFL_PRQ	0x002         /* Request imediate proccessing of frame by receiver station */
#define U2UBFL_ARQ	0x001         /* Request imediate acknowledge by receiving station */

/* Definition of chip_buff usage */
#define ul_usb_in_urb_st chip_buff[0] /* completion status of  ?_complete_in */
#define ul_usb_max_pkt_size chip_buff[1] /* maximal USB packet size */
#define ul_usb_max_msg_inpr chip_buff[2] /* maximal number of messages in progress */
#define ul_usb_outpipe_nr   chip_buff[3] /* OUT pipe number */
#define ul_usb_inpipe_nr    chip_buff[4] /* IN pipe number */


#ifdef FOR_LINUX_KERNEL
#if (LINUX_VERSION_CODE <= VERSION(2,4,10))
  #define usb_fill_bulk_urb FILL_BULK_URB
#endif

#if (LINUX_VERSION_CODE < VERSION(2,5,0))
  #define kc_usb_alloc_urb(urb, gfp) usb_alloc_urb(urb) 
  #define kc_usb_submit_urb(urb, gfp) usb_submit_urb(urb) 
#else
  #define kc_usb_alloc_urb usb_alloc_urb
  #define kc_usb_submit_urb usb_submit_urb
#endif
#endif /*FOR_LINUX_KERNEL*/


#ifdef _WIN32
int ul_usb_msg_inpr=0;	/* number of concurrent messages sent over USB */
#else
extern int ul_usb_msg_inpr;	/* number of concurrent messages sent over USB */
#endif /* _WIN32 */


/* Support for multiple submitted messages */

typedef struct ul_usb_tx_fifo_slot{
  ul_mem_blk    *msg;
  unsigned char ustamp;
}ul_usb_tx_fifo_slot;

typedef struct ul_usb_tx_fifo{
  ul_usb_tx_fifo_slot slot[U2U_MAX_MSG_INPR];
  int size;
  int used;
  int idxin;
  int idxout;
}ul_usb_tx_fifo;

static int ul_usb_tx_fifo_empty(const ul_usb_tx_fifo *fifo)
{
  return fifo->slot[fifo->idxout].msg==NULL;
}

static int ul_usb_tx_fifo_full(const ul_usb_tx_fifo *fifo)
{
  return fifo->slot[fifo->idxin].msg!=NULL;
}

static int ul_usb_tx_fifo_incidx(const ul_usb_tx_fifo *fifo, int idx)
{
  if(++idx>=fifo->size) idx=0;
  return idx;
}

static void ul_usb_tx_fifo_print(const ul_usb_tx_fifo *fifo)
{
  int i;
  LOG_MESSAGES("ul_usb: tx_fifo: idxin=%d idxout=%d [",
            fifo->idxin,fifo->idxout);
  for(i=0;i<fifo->size;i++){
    if(i) LOG_MESSAGES(",");
    if(!fifo->slot[i].msg){
      LOG_MESSAGES("Empty");
    }else{
      LOG_MESSAGES("%d",fifo->slot[i].ustamp);
    }
  }
  LOG_MESSAGES("]\n");
}

/*** Input completion for Petr Smolik's USB2uLan converter ***/
#ifdef FOR_LINUX_KERNEL
#if (LINUX_VERSION_CODE >= VERSION(2,6,0))
  void ul_usb_ps1_complete_in(struct urb *urb,struct pt_regs *regs)
#else
  void ul_usb_ps1_complete_in(struct urb *urb)
#endif
#else /*FOR_LINUX_KERNEL*/
  NTSTATUS ul_usb_ps1_complete_in( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
#endif /*FOR_LINUX_KERNEL*/
{
  ul_drv *udrv;
  int st;
  
  #ifdef FOR_LINUX_KERNEL
    udrv=urb->context;
  #else /*FOR_LINUX_KERNEL*/
    udrv=(ul_drv*)Context;
  #endif /*FOR_LINUX_KERNEL*/
  if(udrv->magic!=UL_DRV_MAGIC){
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_complete_in: Wrong uLan MAGIC number!!!\n");
    #ifdef _WIN32
      return STATUS_INVALID_PARAMETER;
    #else /* _WIN32 */
      return;
    #endif /* _WIN32 */
  }
  #ifdef FOR_LINUX_KERNEL
  if(urb->status<0) 
    st=16-urb->status;
   else st=2;
  #else
    st=2;
  #endif

  LOG_IRQ(KERN_CRIT "ul_usb_ps1_complete_in\n");
  UL_MB();
  udrv->ul_usb_in_urb_st=st;
  uld_kwt_wake(udrv);
  
  #ifdef _WIN32
      return STATUS_SUCCESS;
  #endif /* _WIN32 */
}


/*** Send one frame to USB2uLan converter ***/
int ul_usb_ps1_send_frame(ul_drv *udrv, ul_mem_blk *frame,
                           unsigned char ustamp)
{
  unsigned char buf[U2U_PKT_BUF_SIZE];
  int i, len, ret;
  ul_data_it di;
  struct usb_device *dev=(struct usb_device *)udrv->dev;
    int max_pkt=udrv->ul_usb_max_pkt_size;

  buf[0]=UL_BLK_HEAD(frame).dadr;
  buf[1]=UL_BLK_HEAD(frame).sadr;
  buf[2]=UL_BLK_HEAD(frame).cmd;
  i=UL_BLK_HEAD(frame).flg; buf[3]=0;
  if(i&UL_BFL_ARQ) buf[3]|=U2UBFL_ARQ;
  if(i&UL_BFL_PRQ) buf[3]|=U2UBFL_PRQ;
  if(i&UL_BFL_NORE) buf[3]|=U2UBFL_NORE;
  if(i&UL_BFL_TAIL) buf[3]|=U2UBFL_TAIL;
  if(i&UL_BFL_REC) buf[3]|=U2UBFL_REC;
  buf[4]=ustamp;
  buf[5]=0;
  len=UL_BLK_HEAD(frame).len;
  if(i&UL_BFL_REC){	/* Placeholder message for receive does not */
    len=0;		/* need data, but should check UL_BFL_LNMM !!! */
  }
  buf[6]=(unsigned char)len; buf[7]=(unsigned char)(len/0x100);

  LOG_IRQ(KERN_CRIT "ul_usb : sending head URB (d:%d,s:%d,c:%X,f:%X,l:%d) stmp %d\n",
               buf[0],buf[1],buf[2],buf[3],
               buf[6]+0x100*buf[7],buf[4]);

  ret=usb_bulk_msg(dev, usb_sndbulkpipe(dev, udrv->ul_usb_outpipe_nr), buf, 8, &i, ULD_HZ);
  if(ret<0) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_send_frame: usb_bulk_write head error %d returned\n",ret);
    return ret;
  }

  ul_di_init(&di,frame);
  while(len>0) {
    i=len<max_pkt?len:max_pkt; len-=i;
    ul_di_read(&di, buf, i);
    ret=usb_bulk_msg(dev, usb_sndbulkpipe(dev, udrv->ul_usb_outpipe_nr), buf, i, &i, ULD_HZ);
    if(ret<0) {
      LOG_FATAL(KERN_CRIT "ul_usb_ps1_send_frame: usb_bulk_write data error %d returned\n",ret);
      return ret;
    }
  }
  return di.pos;
}

/*** Receive data of one frame from USB2uLan converter ***/
int ul_usb_ps1_receive_data(ul_drv *udrv, ul_mem_blk *frame, int len)
{
  unsigned char buf[U2U_PKT_BUF_SIZE];
  int i, ret;
  ul_data_it di;
  struct usb_device *dev=(struct usb_device *)udrv->dev;
  int max_pkt=udrv->ul_usb_max_pkt_size;
  
  ul_di_init(&di,frame);
  while(len>0) {
    i=len<max_pkt?len:max_pkt; len-=i;
    ret=usb_bulk_msg(dev, usb_rcvbulkpipe(dev, udrv->ul_usb_inpipe_nr), buf, i, &i, ULD_HZ);
    if(ret<0) {
      LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: usb_bulk_read data error %d returned\n",ret);
      return ret;
      break;
    } else {
      if(i!=ul_di_write(&di, buf, i)){
        #ifdef FOR_LINUX_KERNEL
	  return -EIO;
        #else
          return -1;
        #endif /* FOR_LINUX_KERNEL */
      }
    }
  }
  return di.pos;
}


/*** Worker thread for Petr Smolik's USB2uLan converter ***/
int ul_usb_ps1_thread(void *ptr)
{
  ul_drv *udrv=(ul_drv *)ptr;
  struct usb_device *dev;
  int ret;
  int i, len;
  int fatal_err=0;
  volatile unsigned char in_buf[8];
  unsigned char u2u_ret[2];
  unsigned char ustamp=0;
  ul_mem_blk *tx_mes=NULL;
  ul_mem_blk *rx_mes=NULL;
  ul_mem_blk *frame;
  ul_mem_blk *tail_frame=NULL;
  long tx_sta_time;
  ul_usb_tx_fifo tx_fifo;
  #ifdef FOR_LINUX_KERNEL
    struct urb *in_urb;
  #else
    PURB in_urb=NULL;
    PIRP in_irp=NULL;
  #endif /* FOR_LINUX_KERNEL */
  volatile unsigned in_urb_st;

  if(udrv->magic!=UL_DRV_MAGIC)
  {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: Wrong uLan MAGIC number!!!\n");
    uld_kwt_set_stopped(udrv);
    return -ENODEV;
  }
  dev=(struct usb_device *)udrv->dev;
  if(udrv->ul_usb_max_pkt_size>U2U_PKT_BUF_SIZE)
    udrv->ul_usb_max_pkt_size=U2U_PKT_BUF_SIZE;

  if(udrv->ul_usb_max_msg_inpr>U2U_MAX_MSG_INPR)
    udrv->ul_usb_max_msg_inpr=U2U_MAX_MSG_INPR;
  memset(&tx_fifo,0,sizeof(tx_fifo));
  tx_fifo.size=udrv->ul_usb_max_msg_inpr;

#ifdef FOR_LINUX_KERNEL
#if (LINUX_VERSION_CODE >= VERSION(2,6,0))
  daemonize("kulusbps1d");
  /*allow_signal(SIGKILL);*/
#else /* <2.6.0 */
  lock_kernel();
  daemonize();

 #if (LINUX_VERSION_CODE >= VERSION(2,4,10))
  exit_files(current);
  current->files = init_task.files;
  atomic_inc(&current->files->count);
  daemonize();
  reparent_to_init();
 #endif /* >=2.4.10 */

  /* avoid getting signals */
  spin_lock_irq(&current->sigmask_lock);
  flush_signals(current);
  sigfillset(&current->blocked);
  recalc_sigpending(current);
  spin_unlock_irq(&current->sigmask_lock);

  set_fs(KERNEL_DS);
  strcpy(current->comm, "kulusbps1d");
  unlock_kernel();
#endif /* <2.6.0 */
#endif /*FOR_LINUX_KERNEL*/

  if(udrv->baud_val)
    udrv->baud_div=(udrv->baud_base+udrv->baud_val/2)/udrv->baud_val;

  /* Set converter parameters */
  ret=usb_control_msg(dev, usb_rcvctrlpipe(dev,0), 
                      VENDOR_START_ULAN, USB_TYPE_VENDOR,
                      /*value*/ (short)udrv->baud_val, /*index*/ (short)udrv->my_adr, u2u_ret, 1, 2*ULD_HZ);
  
  if(ret<0) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: KWT usb_control_msg error %d returned\n",ret);
    fatal_err=1;
  }

  if(!u2u_ret[0]) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: not supported baudrate (%d)\n",udrv->baud_val);
    fatal_err=1;
  }
 
  #ifdef FOR_LINUX_KERNEL
    if((in_urb=kc_usb_alloc_urb(0, GFP_KERNEL))) {
      /* URB allocated OK */
    }else{
      LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: KWT cannot alloc in_urb\n");
      fatal_err=1;
    }
  #else
    in_urb = MALLOC(sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER));
    if ( !in_urb) {
      LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: KWT cannot alloc in_urb\n");
      fatal_err=1;
    }
  #endif /* FOR_LINUX_KERNEL */

  usb_clear_halt(dev, usb_sndbulkpipe(dev, udrv->ul_usb_outpipe_nr));
  usb_clear_halt(dev, usb_rcvbulkpipe(dev, udrv->ul_usb_inpipe_nr));
  udrv->ul_usb_in_urb_st=0;

  while(!uld_atomic_test_dfl(udrv,KWTKILL) && !fatal_err){
    uld_kwt_confirm(udrv);
    
    while(!ul_usb_tx_fifo_full(&tx_fifo)){ /* when free Tx slot */

      /* Initiate sending of message from the prep_bll queue */
      if(!(tx_mes=udrv->con_message=udrv->prep_bll.first)) break;

      /* printudrvbll(udrv); */
      ul_bll_move_mes(&udrv->work_bll,tx_mes);
      LOG_MESSAGES(KERN_INFO "ul_usb: KWT processing message\n");
      
      if(!++ustamp) ustamp=1;

      /* send frames of message */
      for(frame=tx_mes;frame;frame=UL_BLK_HEAD(frame).next){
	ret=ul_usb_ps1_send_frame(udrv, frame, ustamp);
	if(ret<0) {
          usb_clear_halt(dev, usb_sndbulkpipe(dev, udrv->ul_usb_outpipe_nr));
          UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
          ul_bll_move_mes(&udrv->proc_bll,tx_mes);
          SCHEDULE_BH(udrv);
          /* ulan_do_bh(udrv); */
	  tx_mes=udrv->con_message=NULL;
          break;
	}
	/* Message has no more frames */
	if(!(UL_BLK_HEAD(frame).flg&UL_BFL_TAIL)||
	   (UL_BLK_HEAD(frame).next==NULL)){
	  tx_fifo.slot[tx_fifo.idxin].ustamp=ustamp;
	  tx_fifo.slot[tx_fifo.idxin].msg=tx_mes;
          tx_fifo.idxin=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxin);
          #ifdef FOR_LINUX_KERNEL
            tx_sta_time=jiffies;
          #endif /* FOR_LINUX_KERNEL */
          ul_usb_tx_fifo_print(&tx_fifo);
	}
      }
    }
    
    UL_MB();
    in_urb_st=*(volatile unsigned *)&(udrv->ul_usb_in_urb_st);
    UL_MB();
    if(in_urb_st==2) {
      UL_MB();
      LOG_IRQ(KERN_CRIT "ul_usb : received head URB (d:%d,s:%d,c:%X,f:%X,l:%d) stmp %d\n",
                   in_buf[0],in_buf[1],in_buf[2],in_buf[3],
                   in_buf[6]+0x100*in_buf[7],in_buf[4]);
      if(in_buf[3]&(U2UBFL_PROC|U2UBFL_FAIL)){
        tx_mes=NULL;
        ul_usb_tx_fifo_print(&tx_fifo);
        if(!ul_usb_tx_fifo_empty(&tx_fifo)){
          if(in_buf[4]==tx_fifo.slot[tx_fifo.idxout].ustamp){
	    tx_mes=tx_fifo.slot[tx_fifo.idxout].msg;
	  }else{
            i=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
	    while((i!=tx_fifo.idxout)&&(tx_fifo.slot[i].msg!=NULL)){
	      if(in_buf[4]==tx_fifo.slot[i].ustamp){
	        do{
                  tx_mes=tx_fifo.slot[tx_fifo.idxout].msg;
	          UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
                  ul_bll_move_mes(&udrv->proc_bll,tx_mes);
		  tx_fifo.slot[tx_fifo.idxout].msg=NULL;
		  tx_fifo.idxout=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
                }while(tx_fifo.idxout!=i);
                SCHEDULE_BH(udrv);
	        tx_mes=tx_fifo.slot[tx_fifo.idxout].msg;
              }
              i=ul_usb_tx_fifo_incidx(&tx_fifo,i);
            }
	    if(!tx_mes){
              LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: incorrect ustamp of returned message\n");
	    }else{
              LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: some lost ustamp messages\n");
	    }
          }
	}else{
	  LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: tx_mes for returned message\n");
	}

	if(!tx_mes){
          usb_clear_halt(dev, usb_rcvbulkpipe(dev, udrv->ul_usb_inpipe_nr));
	}else{
	  if(tail_frame&&(UL_BLK_HEAD(tx_mes).flg&UL_BFL_TAIL)){
	    frame=tail_frame;
	    LOG_MESSAGES("ul_usb_ps1_thread: tail receiption\n");
	  }else{
	    frame=tx_mes;
	  }

          if(in_buf[3]&U2UBFL_FAIL){
	    UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
	  }else if(in_buf[3]&U2UBFL_REC) {
	    if(!(UL_BLK_HEAD(frame).flg&UL_BFL_REC)){
	      UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
              LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: tail rec mitchmatch\n");
	    }else{
	      len=in_buf[6]+in_buf[7]*0x100;
	      UL_BLK_HEAD(frame).dadr=in_buf[0];
	      UL_BLK_HEAD(frame).sadr=in_buf[1];
	      UL_BLK_HEAD(frame).cmd=in_buf[2];
	      UL_BLK_HEAD(frame).len=len;
              ret=ul_usb_ps1_receive_data(udrv, frame, len);
	      if(ret<0) {
		UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
        	LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: tail receive problem\n");
	      }
	    }
	  }
	    
	  if((in_buf[3]&U2UBFL_TAIL)&&(UL_BLK_HEAD(frame).flg&UL_BFL_TAIL)
	      &&!(UL_BLK_HEAD(tx_mes).flg&UL_BFL_FAIL)){
	    tail_frame=UL_BLK_HEAD(frame).next;
	    LOG_MESSAGES("ul_usb_ps1_thread: preparing for tail receive\n");
          }else{
	    if(UL_BLK_HEAD(tx_mes).flg&UL_BFL_M2IN || UL_BLK_HEAD(tx_mes).flg&UL_BFL_FAIL ||
	     UL_BLK_HEAD(tx_mes).dadr==udrv->my_adr || !UL_BLK_HEAD(tx_mes).dadr){
              ul_bll_move_mes(&udrv->proc_bll,tx_mes);
              SCHEDULE_BH(udrv);
              /* ulan_do_bh(udrv); */
	    }else{
              ul_bll_free_mes(tx_mes);
	    }
	    tx_mes=NULL;
	    tail_frame=NULL;
	    tx_fifo.slot[tx_fifo.idxout].msg=NULL;
	    tx_fifo.idxout=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
	    ul_usb_tx_fifo_print(&tx_fifo);
	  }
	}
      } else {
	LOG_MESSAGES("uLan : begin of receive message\n");
	i=0;
        /* if(in_buf[3]&U2UBFL_TAIL ) i|=UL_BFL_TAIL; */
        /* if(in_buf[3]&U2UBFL_REC) i|=UL_BFL_REC; */
        rx_mes=ul_new_frame_head(udrv,in_buf[0]/*dadr*/,in_buf[1]/*sadr*/, 
             	                 in_buf[2]/*cmd*/, i/*flg*/);
	if(rx_mes==NULL){ 
	  LOG_FAILS(KERN_ERR "uLan : NO memory for receive\n");
          usb_clear_halt(dev, usb_rcvbulkpipe(dev, udrv->ul_usb_inpipe_nr));
	}else{
	  ul_bll_ins(&udrv->work_bll,rx_mes);
	  len=in_buf[6]+in_buf[7]*0x100;

          ret=ul_usb_ps1_receive_data(udrv, rx_mes, len);

	  if(ret<0) {
            LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: message receive problem\n");
            ul_bll_free_mes(rx_mes);
	  }else{
	    UL_BLK_HEAD(rx_mes).len=len;
            UL_BLK_HEAD(rx_mes).stamp=ul_gen_stamp();
            ul_bll_move_mes(&udrv->proc_bll,rx_mes);
	  }

          rx_mes=NULL;
          SCHEDULE_BH(udrv);
          /* ulan_do_bh(udrv); */
          /* printudrvbll(udrv); */
	}
      }
    }
    
    if(in_urb_st!=1)
    {
      #ifdef FOR_LINUX_KERNEL
        ret=16-udrv->ul_usb_in_urb_st;
        if((ret==-ENODEV)||(ret==-EILSEQ)||(ret==-EPROTO)){
          LOG_FATAL(KERN_INFO "ul_usb: usb2ulan device problem detected\n");
          fatal_err=1;
          continue;
        }
      #endif /* FOR_LINUX_KERNEL */
      *(volatile unsigned *)&(udrv->ul_usb_in_urb_st)=1;
      UL_MB();
      if(uld_atomic_test_dfl(udrv,KWTKILL)) break;
      LOG_IRQ(KERN_INFO "ul_usb: submitting URB\n");
      #ifdef FOR_LINUX_KERNEL
        usb_fill_bulk_urb (in_urb, dev, usb_rcvbulkpipe(dev, udrv->ul_usb_inpipe_nr),
            (void*)in_buf, 8, ul_usb_ps1_complete_in, udrv);
        ret=kc_usb_submit_urb(in_urb, GFP_KERNEL);
        if (ret==-EPIPE) {
          usb_clear_halt(dev, usb_rcvbulkpipe(dev, udrv->ul_usb_inpipe_nr));
          ret=kc_usb_submit_urb(in_urb, GFP_KERNEL);
        }  
        if(ret<0) {
          LOG_FATAL(KERN_INFO "ul_usb: KWT usb_submit_urb fatal error %d\n",ret);
          fatal_err=1;
        }
      #else
        _usb_fill_bulk_urb (in_urb, dev, usb_rcvbulkpipe(dev, udrv->ul_usb_inpipe_nr),
          (void*)in_buf, 8);
        ret=_usb_submit_urb(dev,in_urb, NULL,ul_usb_ps1_complete_in,udrv,&in_irp);
        if ( !NT_SUCCESS(ret)) {
          LOG_FATAL(KERN_INFO "ul_usb: KWT usb_submit_urb fatal error %d\n",ret);
          fatal_err=1;
        }      
      #endif /* FOR_LINUX_KERNEL */

    } else {
      LOG_IRQ(KERN_INFO "ul_usb: KWT sleeping\n");
      uld_kwt_wait(udrv);
      LOG_IRQ(KERN_INFO "ul_usb: KWT wakeup\n");
    }
  }
  /* Set converter parameters */
  ret=usb_control_msg(dev, usb_rcvctrlpipe(dev,0),
                      VENDOR_STOP_ULAN, USB_TYPE_VENDOR,
                      /*value*/ 0, /*index*/ 0, NULL, 0, ULD_HZ);

  if(ret<0) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: KWT usb_control_msg error %d returned\n",ret);
    fatal_err=1;
  }
  
  ul_usb_tx_fifo_print(&tx_fifo);
  while(!ul_usb_tx_fifo_empty(&tx_fifo)){
//    ul_bll_free_mes(tx_fifo.slot[tx_fifo.idxout].msg);
    tx_mes=tx_fifo.slot[tx_fifo.idxout].msg;
    UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
    ul_bll_move_mes(&udrv->proc_bll,tx_mes);
    tx_fifo.slot[tx_fifo.idxout].msg=NULL;
    tx_fifo.idxout=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
  }
  
  if(in_urb) {
    #ifdef FOR_LINUX_KERNEL
      usb_unlink_urb(in_urb);
      usb_free_urb(in_urb);
    #else
      if (in_irp) 
        IoCancelIrp( in_irp);
      FREE( in_urb);
    #endif /* FOR_LINUX_KERNEL */  
  }
  
  #ifdef FOR_LINUX_KERNEL
    lock_kernel();
  #endif /* FOR_LINUX_KERNEL */

  LOG_IRQ(KERN_INFO "ul_usb: KWT finished\n");
  uld_kwt_set_stopped(udrv);

  #ifdef FOR_LINUX_KERNEL
    unlock_kernel();
  #else
    PsTerminateSystemThread( STATUS_SUCCESS);
  #endif /* FOR_LINUX_KERNEL */
  return 0;
}

/*** Wake up worker thread  ***/
int ul_usb_ps1_stroke(ul_drv *udrv)
{
  uld_kwt_wake(udrv);
  return 0;
}

/*** Control functions of chip driver  ***/
int ul_usb_ps1_cctrl(ul_drv *udrv, int ctrl_fnc, int param)
{
  switch (ctrl_fnc)
  {
   case UL_CC_PINIT:	/* initialize ports (parameter baud)*/
  	{ return 0;
  	};
   case UL_CC_PDONE:	/* deinitialize ports */
  	{ return 0;
  	};
   case UL_CC_RQIRQ:	/* request IRQ handling - kernel worker thread*/
        { 
        if(uld_kwt_test_and_set_prepared(udrv)) { 
          LOG_IRQ("ul_usb_ps1_cctrl: KWT already running\n");
          return 0;
        };
	uld_atomic_clear_dfl(udrv,KWTKILL);

        #ifdef FOR_LINUX_KERNEL
        if(kernel_thread(ul_usb_ps1_thread, udrv,
             #if (LINUX_VERSION_CODE < VERSION(2,5,41))
              CLONE_FS | CLONE_FILES | CLONE_SIGNAL
             #else /* 2.5.41 */
              CLONE_KERNEL
             #endif /* 2.5.41 */
            )<0){
  	      LOG_FATAL(KERN_CRIT "ul_usb_ps1_cctrl: Cannot fork KWT thread!!!\n");
  	      /*uld_atomic_clear_dfl(udrv,KWTRUN);*/
	      uld_kwt_set_stopped(udrv);
  	      return UL_RC_EKWT;
          }
        #endif /*FOR_LINUX_KERNEL*/
        #ifdef _WIN32
        {
          NTSTATUS ntStatus;
          ntStatus = PsCreateSystemThread( &udrv->kwt, 
                         THREAD_ALL_ACCESS, NULL, NULL, NULL, ul_usb_ps1_thread, udrv);
          if (!NT_SUCCESS(ntStatus)) { 
  	    LOG_FATAL(KERN_CRIT "ul_usb_ps1_cctrl: Cannot fork KWT thread!!!\n");

  	    /*uld_atomic_clear_dfl(udrv,KWTRUN);*/
	    uld_kwt_set_stopped(udrv);
  	    return UL_RC_EKWT;
          }
        }
        #endif /* _WIN32 */
        LOG_IRQ("ul_usb_ps1_cctrl: KWT forked\n");
        return 0;
        };
   case UL_CC_FREEIRQ:	/* release IRQ handling - kernel worker thread*/
  	{ 
        if(!uld_kwt_is_running(udrv)) { 
          LOG_IRQ("ul_usb_ps1_cctrl: KWT already stopped\n");
          return 0;
      	};
	uld_kwt_kill(udrv);
	uld_kwt_wait_stopped(udrv);
	LOG_IRQ("ul_usb_ps1_cctrl: KWT stopped\n");
  	return 0;
  	};
  };
  return UL_RC_ENOFNC;
};

/*** Petr Smolik's USB2ULAN version 1 initialize ***/
int ul_usb_ps1_init(ul_drv *udrv, int port, int irq, int baud, int options)
{
  udrv->chip_options=options;
  if(!baud) baud=19200;
  udrv->baud_val=baud;
  udrv->baud_base=57600;
  udrv->baud_div=(udrv->baud_base+baud/2)/baud;
  udrv->port=port;
  udrv->irq=0; 
  udrv->fnc_recch=NULL;
  udrv->fnc_sndch=NULL;
  udrv->fnc_wait=NULL;
  udrv->fnc_pool=NULL;
  udrv->fnc_connect=NULL;
  udrv->fnc_cctrl=&ul_usb_ps1_cctrl;
  udrv->fnc_stroke=&ul_usb_ps1_stroke;
  udrv->ul_usb_max_pkt_size=8; /* <= U2U_PKT_BUF_SIZE */
  udrv->ul_usb_max_msg_inpr=ul_usb_msg_inpr;
  if(!udrv->ul_usb_max_msg_inpr)
    udrv->ul_usb_max_msg_inpr=1; /* <= U2U_MAX_MSG_INPR */
  udrv->ul_usb_outpipe_nr=1;     /* OUT pipe number */
  udrv->ul_usb_inpipe_nr=1;      /* IN pipe number */

  return 0;
};


